/*
 * Copyright (c) 2007 The Akuma Project
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 *
 * $Id: hash.c 92 2007-04-28 11:42:36Z eirik $
 */

#include <stdlib.h>
#include <string.h>

#include "alloc.h"
#include "hash.h"

hashtable_t
hash_make(hashvalue size, hashvalue(*hf)(void *, hashvalue),
          int(*cmp)(void *, void *))
{
	hashtable_t ht;

	/* Round up to nearest power of two */
	size -= 1;
#define clp2shift(x) size |= (size >> (x))
	clp2shift(1);
	clp2shift(2);
	clp2shift(4);
	clp2shift(8);
	clp2shift(16);
#undef clp2shift
	size += 1;

	ht = malloc(sizeof *ht);
	if (ht == NULL)
		return NULL;
	ht->entry = malloc(sizeof(*(ht->entry)) * size);
	if (ht->entry == NULL) {
		free(ht);
		return NULL;
	}
	ht->size = size;
	ht->hf = hf;
	ht->cmp = cmp;
	ht->entries = 0;
	memset(ht->entry, '\0', sizeof(*(ht->entry)) * size);

	return ht;
}

void
hash_destroy(hashtable_t ht)
{
	free(ht->entry);
	free(ht);
}

/*
 * Type describes the type of lookup we are doing, finding an empty slot or
 * finding a match.
 *  0: Find empty slot.
 *  1: Find match.
 */
static int
hash_lookup(hashtable_t ht, void *key, int type, hashvalue *slot)
{
	hashvalue hash, starthash;

	starthash = hash = ht->hf(key, ht->size);
	while ((type == 0 && (ht->entry[hash].key != NULL)) ||
	       (type == 1 && (ht->entry[hash].key == NULL ||
	                      ht->cmp(key, ht->entry[hash].key) != 0))) {
		hash = (hash+1) & (ht->size-1);
		if (hash == starthash || ht->entry[hash].key == NULL)
			return 0;
	}

	*slot = hash;
	return 1;
}

void *
hash_get(hashtable_t ht, void *key)
{
	hashvalue hash;

	if (!hash_lookup(ht, key, 1, &hash))
		return NULL;

	return ht->entry[hash].value;
}

int
hash_insert(hashtable_t ht, void *key, void *value)
{
	hashvalue hash;

	if (ht->entries == ht->size)
		return 0;
	if (!hash_lookup(ht, key, 0, &hash))
		return 0;
	ht->entry[hash].key = key;
	ht->entry[hash].value = value;
	ht->entries++;

	return 1;
}

int
hash_remove(hashtable_t ht, void *key)
{
	hashvalue hash;

	if (ht->entries == 0)
		return 0;
	if (!hash_lookup(ht, key, 1, &hash))
		return 0;
	free(ht->entry[hash].key);
	ht->entry[hash].key = NULL;
	ht->entries--;

	return 1;
}

hashvalue
hash_hf_charp(void *buf, hashvalue size)
{
	char *str;
	hashvalue hash;

	for (str = buf, hash = strlen(buf); *str != '\0'; str++) 
		hash = (33*hash)+(*str);
	/* Hash tables are a power of two */
	return (hash & (size-1));
}

int *
hash_int_to_intp(int a)
{
	int *p;

	p = malloc(sizeof(*p));
	*p = a;

	return p;
}

#ifdef TEST_HASH

#include <stdio.h>

int
main(void)
{
	hashtable_t ht;
	hashvalue size = 2;

	ht = hash_make(size, &hash_hf_charp, ((int(*)(void *, void *))&strcmp));
	if (ht == NULL) {
		fprintf(stderr, "No room for hash table\n");
		return 1;
	}

	printf("Added 'jeroen': %d\n", hash_insert(ht, "jeroen", "Foo"));
	printf("Added 'akuma': %d\n", hash_insert(ht, "akuma", "Bar"));
	printf("Added 'eirik': %d\n", hash_insert(ht, "eirik", "Baz"));

	printf("Value of 'jeroen': %p\n", hash_get(ht, "jeroen"));
	printf("Value of 'akuma': %p\n", hash_get(ht, "akuma"));
	printf("Value of 'eirik': %p\n", hash_get(ht, "eirik"));

	printf("Deleted 'akuma': %d\n", hash_remove(ht, "akuma"));
	printf("Value of 'akuma': %p\n", hash_get(ht, "akuma"));
	printf("Deleted 'akuma': %d\n", hash_remove(ht, "akuma"));

	hash_destroy(ht);

	return 0;
}

#endif /* TEST_HASH */

