#include <sys/types.h>
#include <sys/socket.h>

#include <stdio.h>
#include <err.h>

#include "store.h"
#include "protocol.h"
#include "location.h"
#include "alloc.h"
#include "server.h"
#include "poll.h"
#include "user.h"

/* parser must be included first to get YYSTYPE */
#include "protocol-parser.h"
#include "protocol-lexer.h"

int process_command(command_t *c, struct server_request *req);

int
protocol(struct server_request *req)
{
	yyscan_t scanner;
	protocol_parser_t *pstate = protocol_parser_state_init();
	int ret;
	char *buf;

	yylex_init(&scanner);
	yyset_extra(pstate, scanner);
	yy_scan_string(req->buffer, scanner);
	printf("Trying to parse: %s\n", req->buffer);
	yyparse(scanner, pstate);
	yylex_destroy(scanner);
	if (!pstate->errors) {
		ret = process_command(pstate->command, req);
	} else {
		printf("Number of errors: %d\n", pstate->errors);
		ret = 0;
	}

	if (ret) {
		buf = "SUCCESS\n";
	} else {
		buf = "FAILED\n";
	}
	send(req->sock, buf, strlen(buf), 0);

	return ret;
}

int
process_command_update(command_t *c, struct server_request *req)
{
	int ret;

	ret = store_user_location(c->args.update.user, c->args.update.loc);

	if (ret != 0) {
		warnx("storing user location failed :( %d", ret);
	}
	poller_distribute(c->args.update.user, c->args.update.loc);

	return ret != 0 ? 0 : 1;
}

int
process_command_update_user(command_t *c, struct server_request *req)
{
	return process_command_update(c, req);
}

int
process_command_fetch(command_t *c, struct server_request *req)
{
	location_t l;

	if (store_get_location(c->args.user, &l)) {
		warnx("Failed at grabing location for user %s", c->args.user);

		return 0;
	} else {
		printf("Location for %s is:\n", c->args.user);
		location_print(&l);

		return 1;
	}
}

int
process_command_subscribe(command_t *c, struct server_request *req)
{
	return store_add_subscriber(c->args.subscribe.user,
	    c->args.subscribe.subscriber);
}

int
process_command_subscribers(command_t *c, struct server_request *req)
{
	subscriber_cursor_t *cursor;
	char *subscriber;

	cursor = store_get_subscribers(c->args.user);
	if (cursor == NULL)
		return 0;
	
	while ((subscriber = store_get_subscribers_next(cursor))) {
		printf("Subscriber: %s\n", subscriber);
		send(req->sock, subscriber, strlen(subscriber), 0);
		send(req->sock, "\n", 1, 0);
	}
	store_get_subscribers_end(cursor);

	return 1;
}

int
process_command_poll(command_t *c, struct server_request *req)
{
	printf("%s is waiting for updates\n", c->args.user);
	add_poller(c->args.user, req->sock);

	return 1;
}

int
process_command_newuser(command_t *c, struct server_request *req)
{
	int userid;
	char user[128];
	size_t len;

	userid = generate_userid();
	len = snprintf(user, sizeof(user), "%d-%d", server_id, userid);
	printf("Making new user %s\n", user);
	send(req->sock, user, len, 0);
	send(req->sock, "\n", 1, 0);

	return 1;
}

int
process_command(command_t *c, struct server_request *req)
{
	switch (c->type) {
	case CMD_UPDATE_USER:
		return process_command_update_user(c, req);
	case CMD_UPDATE:
		return process_command_update(c, req);
	case CMD_FETCH:
		return process_command_fetch(c, req);
	case CMD_SUBSCRIBERS:
		return process_command_subscribers(c, req);
	case CMD_SUBSCRIBE:
		return process_command_subscribe(c, req);
	case CMD_POLL:
		return process_command_poll(c, req);
	case CMD_NEWUSER:
		return process_command_newuser(c, req);
	default:
		errx(1, "Unknown command type: %d", c->type);
	}
}

protocol_parser_t *
protocol_parser_state_init(void)
{
	protocol_parser_t *pstate = malloc(sizeof(protocol_parser_t));

	pstate->errors = 0;
	pstate->errormsg = NULL;
	pstate->command = NULL;

	return pstate;
}

void
protocol_parser_state_destroy(protocol_parser_t *pstate)
{
	free(pstate);
}

command_t *
command_new(command_type_t type)
{
	command_t *c = malloc(sizeof(*c));

	memset(c, '\0', sizeof(*c));
	c->type = type;

	return c;
}

