#define _GNU_SOURCE
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>

#include <netinet/in.h>

#include <arpa/inet.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <netdb.h>
#include <err.h>
#include <pthread.h>
#include <assert.h>
#include <ctype.h>

#include "protocol.h"
#include "server.h"
#include "alloc.h"
#include "hash.h"

#define BACKLOG 10

fd_set fd_master;
pthread_mutex_t fd_lock;
int fdmax;

hashtable_t clients;

hashvalue
hash_hf_int(void *value, hashvalue size)
{
	int *b;
	hashvalue a;

	b = value;
	a = *b;

	a = (a+0x7ed55d16) + (a<<12);
	a = (a^0xc761c23c) ^ (a>>19);
	a = (a+0x165667b1) + (a<<5);
	a = (a+0xd3a2646c) ^ (a<<9);
	a = (a+0xfd7046c5) + (a<<3);
	a = (a^0xb55a4f09) ^ (a>>16);

	return a & (size-1);
}

int
hash_cmp_int(int *a, int *b)
{
	if (*a < *b)
		return -1;
	else if (*b < *a)
		return 1;
	else
		return 0;
}

int
server_request_handler(struct server_request *req)
{
	char *s = NULL;

	printf("%d\n", req->sock);
	/* XXX: Save data and wait for more */
	s = strchr(req->buffer, '\r');
	if (s == NULL && (s = strchr(req->buffer, '\n')) == NULL)
		errx(1, "Did not get a full line :-(");

	*s = '\0';
	protocol(req);

	s++;
	while (*s == '\n' || *s == '\r')
		s++;

	/* XXX */
	if (*s != '\0')
		warnx("Got more data, but we are discard it...");

	server_connection_release(req->conn);

	return 0;
}

static void *
get_in_addr(struct sockaddr *sa)
{
	if (sa->sa_family == AF_INET)
		return &(((struct sockaddr_in*)sa)->sin_addr);

	return &(((struct sockaddr_in6*)sa)->sin6_addr);
}

int
server_listen(char *port, struct server_handler_args *args)
{
	int sockfd;  // listen on sock_fd, new connection on new_fd
	struct addrinfo hints, *servinfo, *p;
	int yes = 1;
	int rv;

	memset(&hints, 0, sizeof hints);
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE; // use my IP

	if ((rv = getaddrinfo(NULL, port, &hints, &servinfo)) != 0) {
		warn("getaddrinfo: %s", gai_strerror(rv));
		return 0;
	}

	// loop through all the results and bind to the first we can
	for(p = servinfo; p != NULL; p = p->ai_next) {
		if ((sockfd = socket(p->ai_family, p->ai_socktype,
						p->ai_protocol)) == -1) {
			warn("server: socket");
			continue;
		}

		if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &yes,
					sizeof(int)) == -1) {
			warn("setsockopt");
			return 0;
		}

		if (bind(sockfd, p->ai_addr, p->ai_addrlen) == -1) {
			close(sockfd);
			warn("server: bind");
			continue;
		}

		break;
	}

	if (p == NULL)  {
		warn("server: failed to bind\n");
		return 0;
	}

	freeaddrinfo(servinfo); // all done with this structure

	if (listen(sockfd, BACKLOG) == -1) {
		warn("listen");
		return 0;
	}

#if 0
	sa.sa_handler = sigchld_handler; // reap all dead processes
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;
	if (sigaction(SIGCHLD, &sa, NULL) == -1) {
		warn("sigaction");
		return -1;
	}
#endif

	pthread_mutex_lock(&fd_lock);
	FD_ZERO(&fd_master);
	pthread_mutex_unlock(&fd_lock);

	args->listen_socket = sockfd;

	clients = hash_make(64, &hash_hf_int,
	    (int (*)(void *, void *))&hash_cmp_int);

	return 1;
}

void
server_setfd(int fd)
{
	pthread_mutex_lock(&fd_lock);
	FD_SET(fd, &fd_master);
	if (fdmax < fd)
		fdmax = fd;
	pthread_mutex_unlock(&fd_lock);
}

void
server_unsetfd(int fd)
{
	pthread_mutex_lock(&fd_lock);
	FD_CLR(fd, &fd_master);
	pthread_mutex_unlock(&fd_lock);
}

void
server_connection_grab(struct server_connection *conn)
{
	pthread_mutex_lock(conn->lock);
	conn->ref++;
	pthread_mutex_unlock(conn->lock);
}

void
server_connection_release(struct server_connection *conn)
{
	pthread_mutex_t *lock;
	pthread_mutex_lock(conn->lock);
	lock = conn->lock;
	conn->ref--;
	if (conn->ref == 0) {
		close(conn->sock);
		hash_remove(clients, &conn->sock);
		free(conn);
		conn = NULL;
	}
	pthread_mutex_unlock(lock);
	if (conn == NULL)
		free(lock);
}

void *
server_handler(void *arg)
{
	char s[INET6_ADDRSTRLEN];
	struct sockaddr_storage their_addr; // connector's address information
	socklen_t sin_size;
	int new_fd;
	struct server_handler_args *args = arg;
	fd_set fd_read;
	int i;

	server_setfd(args->listen_socket);

	printf("Starting to accept connections...\n");
	for (;;) {
		fd_read = fd_master;
		if (select(fdmax+1, &fd_read, NULL, NULL, NULL) == -1) {
			warn("select");
			continue;
		}

		for (i = 0; i <= fdmax; i++) {
			if (!FD_ISSET(i, &fd_read))
				continue;
			if (i == args->listen_socket) {
				struct server_connection *conn;

				sin_size = sizeof their_addr;
				new_fd = accept(i,
				    (struct sockaddr *)&their_addr, &sin_size);
				if (new_fd == -1) {
					warn("accept");
					goto loop_end;
				}

				conn = malloc(sizeof(*conn));
				conn->lock = malloc(sizeof(*conn->lock));
				conn->sock = new_fd;
				conn->ref = 0;
				pthread_mutex_init(conn->lock, NULL);

				inet_ntop(their_addr.ss_family,
				    get_in_addr((struct sockaddr *)&their_addr),
				    s, sizeof s);
				printf("server: got connection from %s, %d\n", s, new_fd);

				hash_insert(clients, hash_int_to_intp(new_fd),
				    conn);
				server_connection_grab(conn);
				server_setfd(new_fd);
			} else {
				struct server_request *req;
				struct server_connection *conn;

				req = malloc(sizeof(*req));
				req->n = recv(i, req->buffer,
				    sizeof(req->buffer)-1, 0);
				req->sock = i;
				req->buffer[req->n] = '\0';
				conn = hash_get(clients, &i);
				req->conn = conn;
				switch (req->n) {
				case -1:
					warn("recv()");
					break;
				case 0:
					printf("Client disconnected: %d\n", i);
					server_unsetfd(i);
					if (conn)
						server_connection_release(conn);
					else
						warnx("What? no connection data?");
					break;
				default:
					server_connection_grab(conn);
					queue_add(args->queue,
					    (queue_func)server_request_handler,
					    req);
					break;
				}
			}
		}
 loop_end:
		;
	}

	return NULL;
}

struct server_list *server_list;
int n_servers, max_servers;

void
server_read_list(char *file)
{
	FILE *fp;
	char *line;
	size_t len = 0;
	ssize_t read;
	struct server_list *list;

	if ((fp = fopen(file, "r")) == NULL)
		err(1, "Unable to open server list: %s", file);

	max_servers = 16;
	list = calloc(max_servers, sizeof(*list));
	n_servers = 0;
	while((read = getline(&line, &len, fp)) != -1) {
		int id;
		char *p, *p2;

		if (n_servers > max_servers)
			errx(1, "Too many servers in list");

		p = line;
		while (isspace(*p))
			p++;

		if (*p == '\n')
			continue;

		id = atoi(p);
		assert(id > 0);
		id--;
		list[id].id = id;
		p = strchr(p, ' ');
		assert(p);
		p++;
		p2 = strchr(p, ' ');
		assert(p2);
		strncpy(list[id].ip, p, p2-p);
		list[n_servers].ip[p2-p] = '\0';
		p2++;
		p = p2;
		p2 = strchr(p, '\n');
		assert(p2);
		strncpy(list[id].backup_ip, p, p2-p);
		list[id].backup_ip[p2-p] = '\0';

		printf("server: %d %s %s\n", list[id].id,
		    list[id].ip, list[id].backup_ip);

		n_servers++;
	}

	if (line)
		free(line);

	server_list = list;

	fclose(fp);
}

struct server_list *
servet_get(int id)
{
	assert(id > 0 && id <= max_servers);
	assert(server_list[id-1].id > 0);

	return &server_list[id-1];
}
