#include <sys/stat.h>

#include <stdint.h>
#include <stdio.h>
#include <db.h>
#include <string.h>

#include "location.h"
#include "store.h"
#include "alloc.h"

enum db_type {
	DB_LOC,
	DB_SUBSCRIBERS,
};

char *databases[] = {
	"db/loc.db",
	"db/subscribers.db",
};

DB *
store_open_db(int *ret, enum db_type type)
{
	DB *dbp;
	uint32_t flags;
	struct stat st;

	dbp = NULL;
	if ((*ret = db_create(&dbp, NULL, 0)) != 0) {
		return NULL;
	}

	flags = DB_THREAD;
	if (type == DB_SUBSCRIBERS)
		dbp->set_flags(dbp, DB_DUP);
	printf("Opening database: %s\n", databases[type]);
	if (stat(databases[type], &st) == -1) {
		fprintf(stderr, "Creating a new database...\n");
		flags += DB_CREATE;
	}

	*ret = dbp->open(dbp, NULL, databases[type], NULL, DB_BTREE, flags, 0);

	return dbp;
}

void
store_close_db(DB *dbp)
{
	if (dbp != NULL) {
		dbp->close(dbp, 0);
	}
}

int
store_user_location(char *user, location_t *l)
{
	DB *dbp;
	int ret;
	DBT key, data;

	ret = 0;
	dbp = store_open_db(&ret, DB_LOC);
	if (ret != 0)
		goto err;

	printf("And we want to store location for: %s\n", user);
	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));
	key.data = user;
	key.size = strlen(user);
	data.data = l;
	data.size = sizeof(*l);

	ret = dbp->put(dbp, NULL, &key, &data, 0);
	if (ret != 0)
		goto err;

 err:
	if (ret != 0)
		dbp->err(dbp, ret, "foo");
	store_close_db(dbp);

	return ret;
}

int
store_get_location(char *user, location_t *l)
{
	DB *dbp;
	int ret;
	DBT key, data;

	ret = 0;
	dbp = store_open_db(&ret, DB_LOC);
	if (ret != 0)
		goto err;

	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));

	key.data = user;
	key.size = strlen(user);
	data.data = l;
	data.ulen = sizeof(*l);
	data.flags = DB_DBT_USERMEM;

	ret = dbp->get(dbp, NULL, &key, &data, 0);
	if (ret != 0)
		goto err;
	
 err:
	if (ret != 0)
		dbp->err(dbp, ret, "dbp->get");

	store_close_db(dbp);

	return ret;
}

int
store_add_subscriber(char *user, char *subscriber)
{
	DB *dbp;
	int ret;
	DBT key, data;
	DBC *cursor;

	ret = 0;
	cursor = NULL;
	dbp = store_open_db(&ret, DB_SUBSCRIBERS);
	if (ret != 0)
		goto err;

	printf("Adding subscriber %s to %s\n", subscriber, user);

	if (dbp->cursor(dbp, NULL, &cursor, 0) != 0)
		goto err;

	printf("Adding %s as subscribe to %s updates\n", subscriber, user);
	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));
	key.data = user;
	key.size = strlen(user);
	data.data = subscriber;
	data.size = strlen(subscriber)+1; // We want the trailing '\0'

	ret = cursor->put(cursor, &key, &data, DB_KEYFIRST);
	if (ret != 0)
		goto err;

 err:
	if (ret != 0)
		dbp->err(dbp, ret, "bar");
	if (cursor != NULL)
		cursor->close(cursor);
	store_close_db(dbp);

	return ret == 0 ? 1 : 0;
}

subscriber_cursor_t *
store_get_subscribers(char *user)
{
	subscriber_cursor_t *subs;
	DB *dbp;
	int ret;

	ret = 0;
	dbp = store_open_db(&ret, DB_SUBSCRIBERS);
	if (ret != 0)
		goto err;

	subs = malloc(sizeof(*subs));
	subs->num = 0;
	if ((ret = dbp->cursor(dbp, NULL, &subs->cursor, 0)) != 0)
		goto err;
	subs->user = strdup(user);
	memset(&subs->key, '\0', sizeof(subs->key));
	memset(&subs->data, '\0', sizeof(subs->data));

	printf("Looking for subscriber for: %s\n", user);
	subs->key.data = user;
	subs->key.size = strlen(user);
	subs->data.data = subs->subscriber;
	subs->data.ulen = sizeof(subs->subscriber);
	subs->data.flags = DB_DBT_USERMEM;
	
 err:
	return ret > 0 ? NULL : subs;
}

char *
store_get_subscribers_next(subscriber_cursor_t *c)
{
	int ret, flag;
	db_recno_t a;
	
	if (c->num == 0)
		flag = DB_SET;
	else
		flag = DB_NEXT_DUP;
	ret = c->cursor->get(c->cursor, &c->key, &c->data, flag);
	if (ret != 0)
		return NULL;
	c->num++;

	return c->subscriber;
}

void
store_get_subscribers_end(subscriber_cursor_t *c)
{
	c->cursor->close(c->cursor);
	free(c->user);
	free(c);
}
